/*
 * "$Id$"
 *
 *   DNS-SD discovery backend for the Common UNIX Printing System (CUPS).
 *
 *   Copyright 2008-2009 by Apple Inc.
 *
 *   These coded instructions, statements, and computer programs are the
 *   property of Apple Inc. and are protected by Federal copyright
 *   law.  Distribution and use rights are outlined in the file "LICENSE.txt"
 *   "LICENSE" which should have been included with this file.  If this
 *   file is missing or damaged, see the license at "http://www.cups.org/".
 *
 *   This file is subject to the Apple OS-Developed Software exception.
 *
 * Contents:
 *
 *   main()                  - Browse for printers.
 *   browse_callback()       - Browse devices.
 *   browse_local_callback() - Browse local devices.
 *   compare_devices()       - Compare two devices.
 *   exec_backend()          - Execute the backend that corresponds to the
 *                             resolved service name.
 *   get_device()            - Create or update a device.
 *   query_callback()        - Process query data from DNS-SD
 *   find_device()	     - Process query data.
 *   unquote()               - Unquote a name string.
 */

/*
 * Include necessary headers.
 */

#include "backend-private.h"
#include <cups/array.h>
#ifdef HAVE_AVAHI
# include <avahi-client/client.h>
# include <avahi-client/lookup.h>
# include <avahi-common/simple-watch.h>
# include <avahi-common/domain.h>
# include <avahi-common/error.h>
#define kDNSServiceMaxDomainName AVAHI_DOMAIN_NAME_MAX
#else
# include <dns_sd.h>
#endif /* Avahi */


/*
 * Device structure...
 */

typedef enum
{
  CUPS_DEVICE_PRINTER = 0,		/* lpd://... */
  CUPS_DEVICE_IPP,			/* ipp://... */
  CUPS_DEVICE_FAX_IPP,			/* ipp://... */
  CUPS_DEVICE_PDL_DATASTREAM,		/* socket://... */
  CUPS_DEVICE_RIOUSBPRINT		/* riousbprint://... */
} cups_devtype_t;


typedef struct
{
#ifdef HAVE_DNSSD
  DNSServiceRef	ref;			/* Service reference for resolve */
#endif /* HAVE_DNSSD */
  char		*name,			/* Service name */
		*domain,		/* Domain name */
		*fullName,		/* Full name */
		*make_and_model;	/* Make and model from TXT record */
  cups_devtype_t type;			/* Device registration type */
  int		priority,		/* Priority associated with type */
		cups_shared,		/* CUPS shared printer? */
		sent;			/* Did we list the device? */
} cups_device_t;


/*
 * Local functions...
 */

#ifdef HAVE_AVAHI
/*
 * Avahi callback functions
 */
static void	avahi_client_callback(AvahiClient *client,
				      AvahiClientState state,
				      void *context);
static void	avahi_browse_callback(AvahiServiceBrowser *browser,
				      AvahiIfIndex interface,
				      AvahiProtocol protocol,
				      AvahiBrowserEvent event,
				      const char *serviceName,
				      const char *regtype,
				      const char *replyDomain,
				      AvahiLookupResultFlags flags,
				      void *context);
#else
/*
 * libdns_sd callback functions
 */
static void		browse_callback(DNSServiceRef sdRef,
			                DNSServiceFlags flags,
				        uint32_t interfaceIndex,
				        DNSServiceErrorType errorCode,
				        const char *serviceName,
				        const char *regtype,
				        const char *replyDomain, void *context);
static void		browse_local_callback(DNSServiceRef sdRef,
					      DNSServiceFlags flags,
					      uint32_t interfaceIndex,
					      DNSServiceErrorType errorCode,
					      const char *serviceName,
					      const char *regtype,
					      const char *replyDomain,
					      void *context);
static void		query_callback(DNSServiceRef sdRef,
			               DNSServiceFlags flags,
				       uint32_t interfaceIndex,
				       DNSServiceErrorType errorCode,
				       const char *fullName, uint16_t rrtype,
				       uint16_t rrclass, uint16_t rdlen,
				       const void *rdata, uint32_t ttl,
				       void *context);
#endif

static cups_device_t *find_device (cups_array_t *devices,
				   cups_device_t *key,
				   const char *priority, size_t priority_len,
				   const char *mfg, size_t mfg_len,
				   const char *mdl, size_t mdl_len,
				   const char *product, size_t product_len,
				   const char *ty, size_t ty_len,
				   const char *printer_type,
				   size_t printer_type_len);
static int		compare_devices(cups_device_t *a, cups_device_t *b);
static void		exec_backend(char **argv);
static cups_device_t	*get_device(cups_array_t *devices,
			            const char *serviceName,
			            const char *regtype,
				    const char *replyDomain);
static void		unquote(char *dst, const char *src, size_t dstsize);
static int		device_type(const char *regtype);


#ifdef HAVE_AVAHI
static AvahiSimplePoll *simple_poll = NULL;
static int avahi_got_callback;
#endif /* HAVE_AVAHI */


/*
 * 'main()' - Browse for printers.
 */

int					/* O - Exit status */
main(int  argc,				/* I - Number of command-line args */
     char *argv[])			/* I - Command-line arguments */
{
  const char	*name;			/* Backend name */
  int		fd;			/* Main file descriptor */
  fd_set	input;			/* Input set for select() */
  struct timeval timeout;		/* Timeout for select() */
  cups_array_t	*devices;		/* Device array */
  cups_device_t	*device;		/* Current device */
  char		uriName[1024];		/* Unquoted fullName for URI */
#ifdef HAVE_AVAHI
  AvahiClient	*client;
  int		error;
#else
  DNSServiceRef	main_ref,		/* Main service reference */
		fax_ipp_ref,		/* IPP fax service reference */
		ipp_ref,		/* IPP service reference */
		ipp_tls_ref,		/* IPP w/TLS service reference */
		local_fax_ipp_ref,	/* Local IPP fax service reference */
		local_ipp_ref,		/* Local IPP service reference */
		local_ipp_tls_ref,	/* Local IPP w/TLS service reference */
		local_printer_ref,	/* Local LPD service reference */
		pdl_datastream_ref,	/* AppSocket service reference */
		printer_ref,		/* LPD service reference */
		riousbprint_ref;	/* Remote IO service reference */
#endif /* !HAVE_AVAHI */


 /*
  * Check command-line...
  */

  setbuf(stderr, NULL);

  if (argc >= 6)
    exec_backend(argv);
  else if (argc != 1)
  {
    fprintf(stderr, "Usage: %s job user title copies options [filename(s)]\n",
            argv[0]);
    return (1);
  }

 /*
  * Only do discovery when run as "dnssd"...
  */

  if ((name = strrchr(argv[0], '/')) != NULL)
    name ++;
  else
    name = argv[0];

  if (strcmp(name, "dnssd"))
    return (0);

 /*
  * Create an array to track devices...
  */

  devices = cupsArrayNew((cups_array_func_t)compare_devices, NULL);

 /*
  * Browse for different kinds of printers...
  */

#ifdef HAVE_AVAHI
  if ((simple_poll = avahi_simple_poll_new ()) == NULL)
  {
    perror ("ERROR: Unable to create avahi simple poll object");
    return (1);
  }

  client = avahi_client_new (avahi_simple_poll_get (simple_poll),
			     0, avahi_client_callback, NULL, &error);
  if (!client)
  {
    perror ("ERROR: Unable to create avahi client");
    return (1);
  }

  avahi_service_browser_new (client, AVAHI_IF_UNSPEC,
			     AVAHI_PROTO_UNSPEC,
			     "_fax-ipp._tcp", NULL, 0,
			     avahi_browse_callback, devices);
  avahi_service_browser_new (client, AVAHI_IF_UNSPEC,
			     AVAHI_PROTO_UNSPEC,
			     "_ipp._tcp", NULL, 0,
			     avahi_browse_callback, devices);
  avahi_service_browser_new (client, AVAHI_IF_UNSPEC,
			     AVAHI_PROTO_UNSPEC,
			     "_ipp-tls._tcp", NULL, 0,
			     avahi_browse_callback, devices);
  avahi_service_browser_new (client, AVAHI_IF_UNSPEC,
			     AVAHI_PROTO_UNSPEC,
			     "_pdl-datastream._tcp",
			     NULL, 0,
			     avahi_browse_callback,
			     devices);
  avahi_service_browser_new (client, AVAHI_IF_UNSPEC,
			     AVAHI_PROTO_UNSPEC,
			     "_printer._tcp", NULL, 0,
			     avahi_browse_callback, devices);
  avahi_service_browser_new (client, AVAHI_IF_UNSPEC,
			     AVAHI_PROTO_UNSPEC,
			     "_riousbprint._tcp", NULL, 0,
			     avahi_browse_callback, devices);
#else
  if (DNSServiceCreateConnection(&main_ref) != kDNSServiceErr_NoError)
  {
    perror("ERROR: Unable to create service connection");
    return (1);
  }

  fd = DNSServiceRefSockFD(main_ref);

  fax_ipp_ref = main_ref;
  DNSServiceBrowse(&fax_ipp_ref, kDNSServiceFlagsShareConnection, 0,
                   "_fax-ipp._tcp", NULL, browse_callback, devices);

  ipp_ref = main_ref;
  DNSServiceBrowse(&ipp_ref, kDNSServiceFlagsShareConnection, 0,
                   "_ipp._tcp", NULL, browse_callback, devices);

  ipp_tls_ref = main_ref;
  DNSServiceBrowse(&ipp_tls_ref, kDNSServiceFlagsShareConnection, 0,
                   "_ipp-tls._tcp", NULL, browse_callback, devices);

  local_fax_ipp_ref = main_ref;
  DNSServiceBrowse(&local_fax_ipp_ref, kDNSServiceFlagsShareConnection,
                   kDNSServiceInterfaceIndexLocalOnly,
		   "_fax-ipp._tcp", NULL, browse_local_callback, devices);

  local_ipp_ref = main_ref;
  DNSServiceBrowse(&local_ipp_ref, kDNSServiceFlagsShareConnection,
                   kDNSServiceInterfaceIndexLocalOnly,
		   "_ipp._tcp", NULL, browse_local_callback, devices);

  local_ipp_tls_ref = main_ref;
  DNSServiceBrowse(&local_ipp_tls_ref, kDNSServiceFlagsShareConnection,
                   kDNSServiceInterfaceIndexLocalOnly,
                   "_ipp-tls._tcp", NULL, browse_local_callback, devices);

  local_printer_ref = main_ref;
  DNSServiceBrowse(&local_printer_ref, kDNSServiceFlagsShareConnection,
                   kDNSServiceInterfaceIndexLocalOnly,
                   "_printer._tcp", NULL, browse_local_callback, devices);

  pdl_datastream_ref = main_ref;  
  DNSServiceBrowse(&pdl_datastream_ref, kDNSServiceFlagsShareConnection, 0,
                   "_pdl-datastream._tcp", NULL, browse_callback, devices);

  printer_ref = main_ref;
  DNSServiceBrowse(&printer_ref, kDNSServiceFlagsShareConnection, 0,
                   "_printer._tcp", NULL, browse_callback, devices);

  riousbprint_ref = main_ref;
  DNSServiceBrowse(&riousbprint_ref, kDNSServiceFlagsShareConnection, 0,
                   "_riousbprint._tcp", NULL, browse_callback, devices);
#endif /* !HAVE_AVAHI */

 /*
  * Loop until we are killed...
  */

  for (;;)
  {
    int announce = 0;

#ifdef HAVE_AVAHI
    int r;
    avahi_got_callback = 0;
    r = avahi_simple_poll_iterate (simple_poll, 1);
    if (r != 0 && r != EINTR)
    {
     /*
      * We've been told to exit the loop.  Perhaps the connection to
      * avahi failed.
      */

      break;
    }

    if (avahi_got_callback)
      announce = 1;
#else
    FD_ZERO(&input);
    FD_SET(fd, &input);

    timeout.tv_sec  = 1;
    timeout.tv_usec = 0;

    if (select(fd + 1, &input, NULL, NULL, &timeout) < 0)
      continue;

    if (FD_ISSET(fd, &input))
    {
     /*
      * Process results of our browsing...
      */

      DNSServiceProcessResult(main_ref);
    }
    else
    {
      announce = 1;
    }
#endif /* !HAVE_AVAHI */

    if (announce)
    {
     /*
      * Announce any devices we've found...
      */

#ifndef HAVE_AVAHI
      DNSServiceErrorType status;	/* DNS query status */
#endif /* !HAVE_AVAHI */
      cups_device_t *best;		/* Best matching device */
      char	device_uri[1024];	/* Device URI */
      int	count;			/* Number of queries */
      static const char * const schemes[] =
      		{ "lpd", "ipp", "ipp", "socket", "riousbprint" };
					/* URI schemes for devices */


      for (device = (cups_device_t *)cupsArrayFirst(devices),
               best = NULL, count = 0;
           device;
	   device = (cups_device_t *)cupsArrayNext(devices))
#ifndef HAVE_AVAHI
        if (!device->ref && !device->sent)
	{
	 /*
	  * Found the device, now get the TXT record(s) for it...
	  */

          if (count < 10)
	  {
	    device->ref = main_ref;

	    fprintf(stderr, "DEBUG: Querying \"%s\"...\n", device->fullName);

	    status = DNSServiceQueryRecord(&(device->ref),
				           kDNSServiceFlagsShareConnection,
				           0, device->fullName,
					   kDNSServiceType_TXT,
				           kDNSServiceClass_IN, query_callback,
				           devices);
            if (status != kDNSServiceErr_NoError)
	    {
	      fputs("ERROR: Unable to query for TXT records!\n", stderr);
	      fprintf(stderr, "DEBUG: DNSServiceQueryRecord returned %d\n",
	              status);
            }
	    else
	      count ++;
          }
	}
	else
#endif /* !HAVE_AVAHI */

	if (!device->sent)
	{
#ifndef HAVE_AVAHI
	 /*
	  * Got the TXT records, now report the device...
	  */

	  DNSServiceRefDeallocate(device->ref);
	  device->ref = 0;
#endif /* !HAVE_AVAHI */

          if (!best)
	    best = device;
	  else if (strcasecmp(best->name, device->name) ||
	           strcasecmp(best->domain, device->domain))
          {
	    unquote(uriName, best->fullName, sizeof(uriName));
	    httpAssembleURI(HTTP_URI_CODING_ALL, device_uri, sizeof(device_uri),
			    schemes[best->type], NULL, uriName, 0,
			    best->cups_shared ? "/cups" : "/");
	    cupsBackendReport("network", device_uri, best->make_and_model,
	                      best->name, NULL, NULL);
	    best->sent = 1;
	    best       = device;

#ifdef HAVE_AVAHI
	    /* Hack so that cups-deviced doesn't miss anything (STR #3108). */
	    sleep (1);
#endif
	  }
	  else if (best->priority > device->priority ||
	           (best->priority == device->priority &&
		    best->type < device->type))
          {
	    best->sent = 1;
	    best       = device;
	  }
	  else
	    device->sent = 1;
        }

      if (best)
      {
	unquote(uriName, best->fullName, sizeof(uriName));
	httpAssembleURI(HTTP_URI_CODING_ALL, device_uri, sizeof(device_uri),
			schemes[best->type], NULL, uriName, 0,
			best->cups_shared ? "/cups" : "/");
	cupsBackendReport("network", device_uri, best->make_and_model,
			  best->name, NULL, NULL);
	best->sent = 1;

#ifdef HAVE_AVAHI
	/* Hack so that cups-deviced doesn't miss anything (STR #3108). */
	sleep (1);
#endif
      }
    }
  }
}


#ifdef HAVE_AVAHI
static void
avahi_client_callback(
  AvahiClient *client,
  AvahiClientState state,
  void *context)
{
 /*
  * If the connection drops, quit.
  */

  if (state == AVAHI_CLIENT_FAILURE)
  {
    fprintf (stderr, "ERROR: Avahi connection failed\n");
    avahi_simple_poll_quit (simple_poll);
  }
}

static void
avahi_query_callback(
  AvahiServiceResolver *resolver,
  AvahiIfIndex interface,
  AvahiProtocol protocol,
  AvahiResolverEvent event,
  const char *name,
  const char *type,
  const char *domain,
  const char *host_name,
  const AvahiAddress *address,
  uint16_t port,
  AvahiStringList *txt,
  AvahiLookupResultFlags flags,
  void *context)
{
  AvahiStringList	*pair;
  AvahiClient		*client;
  cups_device_t		key,
			*device;
  char			uqname[1024],
			*ptr;
  char			*priority = NULL,
			*mfg = NULL,
			*mdl = NULL,
			*product = NULL,
			*ty = NULL,
			*printer_type = NULL;
  size_t		priority_len = 0,
			mfg_len = 0,
			mdl_len = 0,
			product_len = 0,
			ty_len = 0,
			printer_type_len = 0;

  client = avahi_service_resolver_get_client (resolver);
  if (event != AVAHI_RESOLVER_FOUND)
  {
    if (event == AVAHI_RESOLVER_FAILURE)
    {
      fprintf (stderr, "ERROR: %s\n",
	       avahi_strerror (avahi_client_errno (client)));
    }

    avahi_service_resolver_free (resolver);
    return;
  }

 /*
  * Set search key for device.
  */

  key.name = uqname;
  unquote (uqname, name, sizeof (uqname));
  if ((ptr = strstr(name, "._")) != NULL)
    *ptr = '\0';

  key.domain = (char *) domain;
  key.type = device_type (type);

 /*
  * Look for information in the TXT string.
  */

  if ((pair = avahi_string_list_find (txt, "priority")) != NULL)
    avahi_string_list_get_pair (pair, NULL, &priority, &priority_len);

  if ((pair = avahi_string_list_find (txt, "usb_MFG")) == NULL)
    pair = avahi_string_list_find (txt, "usb_MANUFACTURER");
  if (pair != NULL)
    avahi_string_list_get_pair (pair, NULL, &mfg, &mfg_len);

  if ((pair = avahi_string_list_find (txt, "usb_MDL")) == NULL)
    pair = avahi_string_list_find (txt, "usb_MODEL");
  if (pair != NULL)
    avahi_string_list_get_pair (pair, NULL, &mdl, &mdl_len);

  if ((pair = avahi_string_list_find (txt, "product")) != NULL)
    avahi_string_list_get_pair (pair, NULL, &product, &product_len);

  if ((pair = avahi_string_list_find (txt, "ty")) != NULL)
    avahi_string_list_get_pair (pair, NULL, &ty, &ty_len);

  if ((pair = avahi_string_list_find (txt, "printer-type")) != NULL)
    avahi_string_list_get_pair (pair, NULL, &printer_type, &printer_type_len);

 /*
  * Find the device and the the TXT information.
  */

  device = find_device ((cups_array_t *) context,
			&key,
			priority, priority_len,
			mfg, mfg_len,
			mdl, mdl_len,
			product, product_len,
			ty, ty_len,
			printer_type, printer_type_len);
  if (device)
  {
   /*
    * Let the main loop know to announce the device.
    */

    avahi_got_callback = 1;
  }
  else
    fprintf (stderr, "DEBUG: Ignoring TXT record for \"%s\"...\n", name);

  avahi_service_resolver_free (resolver);
}

static void
avahi_browse_callback(
  AvahiServiceBrowser *browser,
  AvahiIfIndex interface,
  AvahiProtocol protocol,
  AvahiBrowserEvent event,
  const char *name,
  const char *type,
  const char *domain,
  AvahiLookupResultFlags flags,
  void *context)
{
  AvahiClient *client = avahi_service_browser_get_client (browser);

  switch (event)
  {
  case AVAHI_BROWSER_FAILURE:
    fprintf (stderr, "ERROR: %s\n",
	     avahi_strerror (avahi_client_errno (client)));
    avahi_simple_poll_quit (simple_poll);
    return;

  case AVAHI_BROWSER_NEW:
   /*
    * This object is new on the network.
    */

    if (flags & AVAHI_LOOKUP_RESULT_LOCAL)
    {
     /*
      * This comes from the local machine so ignore it.
      */

      fprintf (stderr, "DEBUG: ignoring local service %s\n", name);
    }
    else
    {
     /*
      * Create a device entry for it if it doesn't yet exist.
      */

      get_device ((cups_array_t *)context, name, type, domain);

     /*
      * Now look for a TXT entry.
      */

      if (avahi_service_resolver_new (client, interface, protocol,
				      name, type, domain,
				      AVAHI_PROTO_UNSPEC, 0,
				      avahi_query_callback, context) == NULL)
      {
	fprintf (stderr, "ERROR: failed to resolve service %s: %s\n",
		 name, avahi_strerror (avahi_client_errno (client)));
      }
    }

    break;

  case AVAHI_BROWSER_REMOVE:
  case AVAHI_BROWSER_ALL_FOR_NOW:
  case AVAHI_BROWSER_CACHE_EXHAUSTED:
    break;
  }  
}

#else /* !HAVE_AVAHI */

/*
 * 'browse_callback()' - Browse devices.
 */

static void
browse_callback(
    DNSServiceRef       sdRef,		/* I - Service reference */
    DNSServiceFlags     flags,		/* I - Option flags */
    uint32_t            interfaceIndex,	/* I - Interface number */
    DNSServiceErrorType errorCode,	/* I - Error, if any */
    const char          *serviceName,	/* I - Name of service/device */
    const char          *regtype,	/* I - Type of service */
    const char          *replyDomain,	/* I - Service domain */
    void                *context)	/* I - Devices array */
{
  fprintf(stderr, "DEBUG2: browse_callback(sdRef=%p, flags=%x, "
                  "interfaceIndex=%d, errorCode=%d, serviceName=\"%s\", "
		  "regtype=\"%s\", replyDomain=\"%s\", context=%p)\n",
          sdRef, flags, interfaceIndex, errorCode,
	  serviceName ? serviceName : "(null)",
	  regtype ? regtype : "(null)",
	  replyDomain ? replyDomain : "(null)",
	  context);

 /*
  * Only process "add" data...
  */

  if (errorCode != kDNSServiceErr_NoError || !(flags & kDNSServiceFlagsAdd))
    return;

 /*
  * Get the device...
  */

  get_device((cups_array_t *)context, serviceName, regtype, replyDomain);
}


/*
 * 'browse_local_callback()' - Browse local devices.
 */

static void
browse_local_callback(
    DNSServiceRef       sdRef,		/* I - Service reference */
    DNSServiceFlags     flags,		/* I - Option flags */
    uint32_t            interfaceIndex,	/* I - Interface number */
    DNSServiceErrorType errorCode,	/* I - Error, if any */
    const char          *serviceName,	/* I - Name of service/device */
    const char          *regtype,	/* I - Type of service */
    const char          *replyDomain,	/* I - Service domain */
    void                *context)	/* I - Devices array */
{
  cups_device_t	*device;		/* Device */


  fprintf(stderr, "DEBUG2: browse_local_callback(sdRef=%p, flags=%x, "
                  "interfaceIndex=%d, errorCode=%d, serviceName=\"%s\", "
		  "regtype=\"%s\", replyDomain=\"%s\", context=%p)\n",
          sdRef, flags, interfaceIndex, errorCode,
	  serviceName ? serviceName : "(null)",
	  regtype ? regtype : "(null)",
	  replyDomain ? replyDomain : "(null)",
	  context);

 /*
  * Only process "add" data...
  */

  if (errorCode != kDNSServiceErr_NoError || !(flags & kDNSServiceFlagsAdd))
    return;

 /*
  * Get the device...
  */

  device = get_device((cups_array_t *)context, serviceName, regtype,
                      replyDomain);

 /*
  * Hide locally-registered devices...
  */

  fprintf(stderr, "DEBUG: Hiding local printer \"%s\"...\n",
	  device->fullName);
  device->sent = 1;
}
#endif /* !HAVE_AVAHI */


/*
 * 'compare_devices()' - Compare two devices.
 */

static int				/* O - Result of comparison */
compare_devices(cups_device_t *a,	/* I - First device */
                cups_device_t *b)	/* I - Second device */
{
  int result = strcmp(a->name, b->name);

  if (result)
    return (result);
  else
    return (strcmp(a->domain, b->domain));
}


/*
 * 'exec_backend()' - Execute the backend that corresponds to the
 *                    resolved service name.
 */

static void
exec_backend(char **argv)		/* I - Command-line arguments */
{
  const char	*resolved_uri,		/* Resolved device URI */
		*cups_serverbin;	/* Location of programs */
  char		scheme[1024],		/* Scheme from URI */
		*ptr,			/* Pointer into scheme */
		filename[1024];		/* Backend filename */


 /*
  * Resolve the device URI...
  */

  if ((resolved_uri = cupsBackendDeviceURI(argv)) == NULL)
    exit(CUPS_BACKEND_FAILED);

 /*
  * Extract the scheme from the URI...
  */

  strlcpy(scheme, resolved_uri, sizeof(scheme));
  if ((ptr = strchr(scheme, ':')) != NULL)
    *ptr = '\0';

 /*
  * Get the filename of the backend...
  */

  if ((cups_serverbin = getenv("CUPS_SERVERBIN")) == NULL)
    cups_serverbin = CUPS_SERVERBIN;

  snprintf(filename, sizeof(filename), "%s/backend/%s", cups_serverbin, scheme);

 /*
  * Overwrite the device URIs and run the new backend...
  */

  setenv("DEVICE_URI", resolved_uri, 1);

  argv[0] = (char *)resolved_uri;

  fprintf(stderr, "DEBUG: Executing backend \"%s\"...\n", filename);

  execv(filename, argv);

  fprintf(stderr, "ERROR: Unable to execute backend \"%s\": %s\n", filename,
          strerror(errno));
  exit(CUPS_BACKEND_STOP);
}


/*
 * 'get_device()' - Create or update a device.
 */

static cups_device_t *			/* O - Device */
get_device(cups_array_t *devices,	/* I - Device array */
           const char   *serviceName,	/* I - Name of service/device */
           const char   *regtype,	/* I - Type of service */
           const char   *replyDomain)	/* I - Service domain */
{
  cups_device_t	key,			/* Search key */
		*device;		/* Device */
  char		fullName[kDNSServiceMaxDomainName];
					/* Full name for query */


 /*
  * See if this is a new device...
  */

  key.name   = (char *)serviceName;
  key.domain = (char *)replyDomain;
  key.type = device_type (regtype);

  for (device = cupsArrayFind(devices, &key);
       device;
       device = cupsArrayNext(devices))
    if (strcasecmp(device->name, key.name) ||
        strcasecmp(device->domain, key.domain))
      break;
    else if (device->type == key.type)
      return (device);

 /*
  * Yes, add the device...
  */

  fprintf(stderr, "DEBUG: Found \"%s.%s%s\"...\n", serviceName, regtype,
	  replyDomain);

  device           = calloc(sizeof(cups_device_t), 1);
  device->name     = strdup(serviceName);
  device->domain   = strdup(replyDomain);
  device->type     = key.type;
  device->priority = 50;

  cupsArrayAdd(devices, device);

 /*
  * Set the "full name" of this service, which is used for queries...
  */

#ifdef HAVE_AVAHI
  avahi_service_name_join (fullName, kDNSServiceMaxDomainName,
			   serviceName, regtype, replyDomain);
  device->fullName = strdup(fullName);
#else
  DNSServiceConstructFullName(fullName, serviceName, regtype, replyDomain);
  device->fullName = strdup(fullName);
#endif /* !HAVE_AVAHI */

  return (device);
}


#ifndef HAVE_AVAHI
/*
 * 'query_callback()' - Process query data.
 */

static void
query_callback(
    DNSServiceRef       sdRef,		/* I - Service reference */
    DNSServiceFlags     flags,		/* I - Data flags */
    uint32_t            interfaceIndex,	/* I - Interface */
    DNSServiceErrorType errorCode,	/* I - Error, if any */
    const char          *fullName,	/* I - Full service name */
    uint16_t            rrtype,		/* I - Record type */
    uint16_t            rrclass,	/* I - Record class */
    uint16_t            rdlen,		/* I - Length of record data */
    const void          *rdata,		/* I - Record data */
    uint32_t            ttl,		/* I - Time-to-live */
    void                *context)	/* I - Devices array */
{
  char		name[1024],		/* Service name */
		*ptr;			/* Pointer into name */
  cups_device_t	key;			/* Search key */
  const char	*priority,
		*mfg,
		*mdl,
		*product,
		*ty,
		*printer_type;
  uint8_t	priority_len,
		mfg_len,
		mdl_len,
		product_len,
		ty_len,
		printer_type_len;

  fprintf(stderr, "DEBUG2: query_callback(sdRef=%p, flags=%x, "
                  "interfaceIndex=%d, errorCode=%d, fullName=\"%s\", "
		  "rrtype=%u, rrclass=%u, rdlen=%u, rdata=%p, ttl=%u, "
		  "context=%p)\n",
          sdRef, flags, interfaceIndex, errorCode,
	  fullName ? fullName : "(null)", rrtype, rrclass, rdlen, rdata, ttl,
	  context);

 /*
  * Only process "add" data...
  */

  if (errorCode != kDNSServiceErr_NoError || !(flags & kDNSServiceFlagsAdd))
    return;

 /*
  * Lookup the service in the devices array.
  */

  key.name = name;

  unquote(name, fullName, sizeof(name));

  if ((key.domain = strstr(name, "._tcp.")) != NULL)
    key.domain += 6;
  else
    key.domain = (char *)"local.";

  if ((ptr = strstr(name, "._")) != NULL)
    *ptr = '\0';

  if (strstr(fullName, "_ipp._tcp.") ||
      strstr(fullName, "_ipp-tls._tcp."))
    key.type = CUPS_DEVICE_IPP;
  else if (strstr(fullName, "_fax-ipp._tcp."))
    key.type = CUPS_DEVICE_FAX_IPP;
  else if (strstr(fullName, "_printer._tcp."))
    key.type = CUPS_DEVICE_PRINTER;
  else if (strstr(fullName, "_pdl-datastream._tcp."))
    key.type = CUPS_DEVICE_PDL_DATASTREAM;
  else
    key.type = CUPS_DEVICE_RIOUSBPRINT;

  priority = TXTRecordGetValuePtr(rdlen, rdata, "priority", &priority_len);
  if ((mfg = TXTRecordGetValuePtr(rdlen, rdata, "usb_MFG", &mfg_len)) == NULL)
    mfg = TXTRecordGetValuePtr(rdlen, rdata, "usb_MANUFACTURER", &mfg_len);

  if ((mdl = TXTRecordGetValuePtr(rdlen, rdata, "usb_MDL", &mdl_len)) == NULL)
    mdl = TXTRecordGetValuePtr(rdlen, rdata, "usb_MODEL", &mdl_len);

  product = TXTRecordGetValuePtr(rdlen, rdata, "product", &product_len);
  ty = TXTRecordGetValuePtr(rdlen, rdata, "ty", &ty_len);
  printer_type = TXTRecordGetValuePtr(rdlen, rdata, "printer-type",
				      &printer_type_len);

  if (!find_device ((cups_array_t *) context,
		    &key,
		    priority, priority_len,
		    mfg, mfg_len,
		    mdl, mdl_len,
		    product, product_len,
		    ty, ty_len,
		    printer_type, printer_type_len))
    fprintf(stderr, "DEBUG: Ignoring TXT record for \"%s\"...\n", fullName);
}
#endif /* !HAVE_AVAHI */


static cups_device_t *
find_device (cups_array_t *devices,
	     cups_device_t *key,
	     const char *priority, size_t priority_len,
	     const char *mfg, size_t mfg_len,
	     const char *mdl, size_t mdl_len,
	     const char *product, size_t product_len,
	     const char *ty, size_t ty_len,
	     const char *printer_type, size_t printer_type_len)
{
  cups_device_t *device;

  for (device = cupsArrayFind(devices, key);
       device;
       device = cupsArrayNext(devices))
  {
    if (strcasecmp(device->name, key->name) ||
        strcasecmp(device->domain, key->domain))
    {
      device = NULL;
      break;
    }
    else if (device->type == key->type)
    {
     /*
      * Found it, pull out the priority and make and model from the TXT
      * record and save it...
      */

      char	make_and_model[512],	/* Manufacturer and model */
		model[256],		/* Model */
		priority_buf[256],	/* Priority */
		*ptr;


      if (priority && priority_len)
      {
	memcpy(priority_buf, priority, priority_len);
	priority_buf[priority_len] = '\0';
	device->priority = atoi(priority_buf);
      }

      if (mfg && mfg_len)
      {
	memcpy(make_and_model, mfg, mfg_len);
	make_and_model[mfg_len] = '\0';
      }
      else
	make_and_model[0] = '\0';

      if (mdl && mdl_len)
      {
	memcpy(model, mdl, mdl_len);
	model[mdl_len] = '\0';
      }
      else if (product && product_len > 2)
      {
	if (product[0] == '(')
	{
	 /*
	  * Strip parenthesis...
	  */

	  memcpy(model, product + 1, product_len - 2);
	  model[product_len - 2] = '\0';
	}
	else
	{
	  memcpy(model, product, product_len);
	  model[product_len] = '\0';
	}

	if (!strcasecmp(model, "GPL Ghostscript") ||
	    !strcasecmp(model, "GNU Ghostscript") ||
	    !strcasecmp(model, "ESP Ghostscript"))
	{
	  if (ty && ty_len)
	  {
	    memcpy(model, ty, ty_len);
	    model[ty_len] = '\0';

	    if ((ptr = strchr(model, ',')) != NULL)
	      *ptr = '\0';
	  }
	  else
	    strcpy(model, "Unknown");
	}
      }
      else
	strcpy(model, "Unknown");

      if (device->make_and_model)
	free(device->make_and_model);

      if (make_and_model[0])
      {
	strlcat(make_and_model, " ", sizeof(make_and_model));
	strlcat(make_and_model, model, sizeof(make_and_model));
	device->make_and_model = strdup(make_and_model);
      }
      else
	device->make_and_model = strdup(model);

      if ((device->type == CUPS_DEVICE_IPP ||
	   device->type == CUPS_DEVICE_PRINTER) &&
	  printer_type)
      {
       /*
	* This is a CUPS printer!
	*/

	device->cups_shared = 1;

	if (device->type == CUPS_DEVICE_PRINTER)
	  device->sent = 1;
      }

      break;
    }
  }

  return device;
}


/*
 * 'unquote()' - Unquote a name string.
 */

static void
unquote(char       *dst,		/* I - Destination buffer */
        const char *src,		/* I - Source string */
	size_t     dstsize)		/* I - Size of destination buffer */
{
  char	*dstend = dst + dstsize - 1;	/* End of destination buffer */


  while (*src && dst < dstend)
  {
    if (*src == '\\')
    {
      src ++;
      if (isdigit(src[0] & 255) && isdigit(src[1] & 255) &&
          isdigit(src[2] & 255))
      {
        *dst++ = ((((src[0] - '0') * 10) + src[1] - '0') * 10) + src[2] - '0';
	src += 3;
      }
      else
        *dst++ = *src++;
    }
    else
      *dst++ = *src ++;
  }

  *dst = '\0';
}


static int
device_type (const char *regtype)
{
#ifdef HAVE_AVAHI
  if (!strcmp(regtype, "_ipp._tcp") ||
      !strcmp(regtype, "_ipp-tls._tcp"))
    return (CUPS_DEVICE_IPP);
  else if (!strcmp(regtype, "_fax-ipp._tcp"))
    return (CUPS_DEVICE_FAX_IPP);
  else if (!strcmp(regtype, "_printer._tcp"))
    return (CUPS_DEVICE_PRINTER);
  else if (!strcmp(regtype, "_pdl-datastream._tcp"))
    return (CUPS_DEVICE_PDL_DATASTREAM);
#else
  if (!strcmp(regtype, "_ipp._tcp.") ||
      !strcmp(regtype, "_ipp-tls._tcp."))
    return (CUPS_DEVICE_IPP);
  else if (!strcmp(regtype, "_fax-ipp._tcp."))
    return (CUPS_DEVICE_FAX_IPP);
  else if (!strcmp(regtype, "_printer._tcp."))
    return (CUPS_DEVICE_PRINTER);
  else if (!strcmp(regtype, "_pdl-datastream._tcp."))
    return (CUPS_DEVICE_PDL_DATASTREAM);
#endif /* !HAVE_AVAHI */

  return (CUPS_DEVICE_RIOUSBPRINT);
}


/*
 * End of "$Id$".
 */
